# -*- coding: utf-8 -*-
"""
Created on Sat Mar 30 12:29:16 2019

@author: Mehdi zamani joharestani
"""
from keras import regularizers
from keras.callbacks import EarlyStopping ,ModelCheckpoint
from keras.models import Sequential
from math import sqrt
from scipy.interpolate import interpn
from sklearn.externals import joblib
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler,MinMaxScaler,RobustScaler
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from keras.layers import Dense


# density_scatter plot function 
def density_scatter( x , y, ax = None, sort = True, bins = 20, figsize=(6,4), **kwargs )   :
    """
    Scatter plot colored by 2d histogram
    """
    x=x.reshape(len(x))
    y=y.reshape(len(y))
    if ax is None :
        fig , ax = plt.subplots(figsize=figsize)
    data , x_e, y_e = np.histogram2d( x, y, bins = bins)
    z = interpn( ( 0.5*(x_e[1:] + x_e[:-1]) , 0.5*(y_e[1:]+y_e[:-1]) ) , data , np.vstack([x,y]).T , method = "splinef2d", bounds_error = False )

    # Sort the points by density, so that the densest points are plotted last
    if sort :
        idx = z.argsort()
        x, y, z = x[idx], y[idx], z[idx]
    cmap = matplotlib.cm.get_cmap('jet')
    normalize = matplotlib.colors.Normalize(vmin=0, vmax=max(z))#min(z)
    ax.scatter( x, y, c=z, **kwargs ,cmap=cmap)
    cax, _ = matplotlib.colorbar.make_axes(ax)
    cbar = matplotlib.colorbar.ColorbarBase(cax, cmap=cmap, norm=normalize)
    z = np.polyfit(x, y, 1)
    p = np.poly1d(z)
    print(z)
    ax.plot(x,p(x),"r-")
    ax.plot(x,x,":",color='grey')
    return ax,fig,z
#======================================================================

# RMSE calculation function
def rmse(y_actual, y_predicted):
    return sqrt(mean_squared_error(y_actual, y_predicted))
#======================================================================


dfte= pd.DataFrame( columns=['features','r2','mae','rmse'])

#======================================================================
step = 'AOD10'# based on alalysis use one of 'NAOD','AOD03','AOD10','AOD310' 

# loading database
file = 'totalnull.csv'  # database  file name          
hint=file[0:-4]+step
df = pd.read_csv(file, na_values=['NA', ' '])
df = df.drop(df.columns[df.columns.str.contains('Unnamed',case = False)],axis = 1) 
lis=['Tehran_12','Fath square', 'Mahallati' ,'Tehran_15','Tehran_17' ]
df=df[ ~df['Station'].isin(lis)]   
#======================================================================

# filtering the missing values based on features of experiment

if step =='NAOD' :
    df_=df_[(df_[ 'PM2.5'] >= 0) &   (df_['PM2.5_lag1']>= 0) & (df_['PM2.5_lag2']>= 0) & (df_['AOD10']>= 0) ].reset_index(drop=True)
elif step =='AOD3' :
    df_=df_[(df_[ 'PM2.5'] >= 0) &   (df_['PM2.5_lag1']>= 0) & (df_['PM2.5_lag2']>= 0) & (df_['AOD03']>= 0)  ].reset_index(drop=True)
elif step =='AOD10' :
    df_=df_[(df_[ 'PM2.5'] >= 0) &   (df_['PM2.5_lag1']>= 0) & (df_['PM2.5_lag2']>= 0) & (df_['AOD10']>= 0)  ].reset_index(drop=True)
elif step =='AOD310' :
    df_=df_[(df_[ 'PM2.5'] >= 0) &   (df_['PM2.5_lag1']>= 0) & (df_['PM2.5_lag2']>= 0) & (df_['AOD10']>= 0) & (df_['AOD03']>= 0)  ].reset_index(drop=True)      
#==============================================================================
    
# X=input features and Y=target selection  
y=df_[['PM2.5']]

LIST=[
'PM2.5_lag1',
'wind_speed',
'visibility ',
'sustained_windsp',
'RH',
'Tmin',
'Altitude',
'T',
'PM2.5_lag2',
'Dayofyear',
'Pressure',
'Tmax',
'Season',
'Weekday',
'Dewpoint',
'AOD10',
'Rainfall_Lag2',
'Distance',
'Latitude',
'Rainfall_Lag1',
'Longitude',
'Rainfall',
'Organizati',
]

Xdataset=df_[LIST]
  

flist= [
         'Dayofyear',
         'Weekday',
         'Season',
         'Latitude',
         'Longitude',
         'Altitude',
         'Organizati',
         'Distance',
         'T',
         'Tmax',
         'Tmin',
         'RH',
         'Rainfall',
         'Rainfall_Lag1',
         'Rainfall_Lag2',
         'visibility ',
         'wind_speed',
         'sustained_windsp',
         'Pressure',
         'Dewpoint',
         'AOD03',
         'AOD10',
         'PM2.5_lag1',
         'PM2.5_lag2',
]
#==============================================================================

# Normalization 
scaler_list=['standard'] #'minmax','robust',
for scaller in scaler_list:
    if scaller== 'standard':
        scalerx = StandardScaler().fit(Xdataset)
        scalery = StandardScaler().fit(y)
    elif scaller == 'robust':
        scalerx = RobustScaler().fit(Xdataset)
        scalery = RobustScaler().fit(y)
    elif scaller == 'minmax':
        scalerx = MinMaxScaler(feature_range=(-1, 1)).fit(Xdataset)
        scalery = MinMaxScaler(feature_range=(-1, 1)).fit(y)
#==============================================================================    
# Scale the train set
    x1 = scalerx.transform(Xdataset)
    y1 = scalery.transform(y)
    dataset_size=len(x1)
    x_train, x_test, y_train, y_test = train_test_split(x1, y1, test_size=0.30, random_state=3)
#==============================================================================
# save scaler
    scaler_filename = './data/deep/scaler/'+scaller+'/Xscaler.save'
    joblib.dump(scalerx, scaler_filename) 
    scaler_filename = './data/deep/scaler/'+scaller+'/Yscaler.save'
    joblib.dump(scalery, scaler_filename) 
 #==============================================================================
# define and fit model             
    Sfilename = './data/deep/scaler/'+scaller+'/totalzeroAOD3100.hdf5'#+hint+'.hdf5'
    print('Build model...')
       
    def create_model():
        model = Sequential()
        model.add(Dense(270, input_dim=x_train.shape[1], activation='relu'))
        model.add(Dense(120, activation='relu',kernel_regularizer=regularizers.l2(0.002)))
        model.add(Dense(70, activation='relu',kernel_regularizer=regularizers.l2(0.002)))
        model.add(Dense(50, activation='relu',kernel_regularizer=regularizers.l2(0.002)))
        model.add(Dense(20, 
                        kernel_regularizer=regularizers.l2(0.001),
                        activity_regularizer=regularizers.l1(0.001),activation='relu'))
        # activity_regularizer can also be used.
        model.add(Dense(1))
        return model
        
    def train(x_train,y_train,x_test,y_test):
       model = create_model()
       opt='adam'#keras.optimizers.Adam(lr=0.01, beta_1=0.9, beta_2=0.999, epsilon=None, decay=0.0, amsgrad=False)
       model.compile(loss='mean_squared_error', optimizer=opt, metrics=['mae'])
       checkpointer = ModelCheckpoint(filepath=Sfilename, verbose=0, save_best_only=True)
       monitor = EarlyStopping(monitor='val_loss', min_delta=1e-6, patience=35, verbose=0, mode='auto')
       history=model.fit(x_train,y_train,validation_data=(x_test,y_test),callbacks=[monitor,checkpointer],verbose=2,epochs=1000)#,batch_size = 70)
       return history ,model
    import datetime
    start_time  = datetime.datetime.now()
    history,model = train(x_train,y_train,x_test,y_test)
    end_time = datetime.datetime.now()
    print ('Select Done..., Time Cost: %d' % ((end_time - start_time).seconds)) 
#==============================================================================        
    
    
# Plot training==================================
    fig0 = plt.figure()
    plt.style.use('bmh')
    plt.plot(history.history['loss'])
    plt.plot(history.history['val_loss'])
    plt.title('model loss')
    plt.ylabel('prediction')
    plt.xlabel('epochs')
    plt.legend(['Train', 'Test'], loc='upper right')
    plt.show()
    plt.savefig('./data/deep/scaler/'+scaller+'/training_'+hint+'.svg')
    plt.savefig('./data/deep/scaler/'+scaller+'/training_'+hint+'.jpg')

# load the best model found during training       
    def load_trained_model(weights_path):
        model = create_model()
        model.load_weights(weights_path)
        model.compile(loss='mean_squared_error', optimizer='adam', metrics=['mae'])
        return model
    model=load_trained_model(Sfilename)
    scalerx = joblib.load("./data/deep/scaler/"+scaller+"/Xscaler.save") 
    scalery = joblib.load("./data/deep/scaler/"+scaller+"/Yscaler.save") 

#======================================================================            
# evaluting the model using test dataset    
    
    y_pred = model.predict(x_test)
    y_predinv = scalery.inverse_transform(y_pred)
    y_testinv = scalery.inverse_transform(y_test)     
    mae=mean_absolute_error(y_testinv, y_predinv)
    r2=r2_score( y_testinv, y_predinv)
    rmse_value= rmse(y_testinv, y_predinv)
    print(' R2  %.3f= '  % r2 )
    print(' MAE %.3f= '  % mae)
    print(' RMSE  %.3f= '  % rmse_value )
    print(r2,'  ',mae,' ', rmse_value)
    data=['No', r2, mae, rmse_value] 
    dfte=dfte.append(pd.DataFrame(  [data]  , columns=['features','r2','mae','rmse']))
#==========================================================================
# ploting the results
    text='R2= '+format(r2, '.2f') + ' Mae= '+format(mae, '.2f')+ ' RMSE= '+format(rmse_value, '.2f')
    ax,fig1,z=density_scatter( y_testinv.reshape(len(y_testinv)), y_predinv, bins = [15,15] , figsize=(10,8))
    ax.set_title('Deep learning Modeling'  , fontsize=16,fontweight='bold')
    pol_text='y=%.2f x+%.2f, DB size=%0.1fK' % (z[0],z[1],(dataset_size/1000))
    ax.set_xlabel("Observed PM2.5", fontsize=16)
    ax.set_ylabel("Predicted PM2.5", fontsize=16)
    ax.legend([pol_text,'y=x ',text], fontsize='x-large')
    ax.tick_params(axis='both', which='major', labelsize=12)
    ax.patch.set_facecolor('white')
    ax.set_ylim(0,250)
    ax.set_xlim(0,250)
#==========================================================================
# featurte permutation

    for par in LIST:  
        dftemp= pd.DataFrame(x_test,columns=LIST)
        dftemp[par] =dftemp[par].apply(lambda x: np.random.choice(dftemp[par], size=1)) 
        x_testnoisy = dftemp
# evalute the model using permuted feature
        y_pred = model.predict(x_testnoisy)
        y_predinv = scalery.inverse_transform(y_pred)
        y_testinv = scalery.inverse_transform(y_test)     
        mae=mean_absolute_error(y_testinv, y_predinv)
        r2=r2_score( y_testinv, y_predinv)
        rmse_value= rmse(y_testinv, y_predinv)
        print(' R2  %.3f= '  % r2 )
        print(' MAE %.3f= '  % mae)
        print(' RMSE  %.3f= '  % rmse_value )
        print(r2,'  ',mae,' ', rmse_value)
        data=[par, r2, mae, rmse_value] 
# recording the evaluation meterics
        dfte=dfte.append(pd.DataFrame(  [data]  , columns=['features','r2','mae','rmse']))
        text='noisy '+par  +' R2= '+format(r2, '.2f') + ' Mae= '+format(mae, '.2f')+ ' RMSE= '+format(rmse_value, '.2f')
        ax,fig1,z=density_scatter( y_testinv.reshape(len(y_testinv)), y_predinv, bins = [15,15] , figsize=(10,8))
        ax.set_title('Deep learning Modeling'  , fontsize=16,fontweight='bold')
        pol_text='y=%.2f x+%.2f, DB size=%0.1fK' % (z[0],z[1],(dataset_size/1000))
        ax.set_xlabel("Observed PM2.5", fontsize=16)
        ax.set_ylabel("Predicted PM2.5", fontsize=16)
        ax.legend([pol_text,'y=x ',text], fontsize='x-large')
        ax.tick_params(axis='both', which='major', labelsize=12)
        ax.patch.set_facecolor('white')
        ax.set_ylim(0,250)
        ax.set_xlim(0,250)
#==========================================================================
#save the results        
dfte.to_csv('deep_result.csv')