# -*- coding: utf-8 -*-
"""
Created on Sat Mar 30 12:29:16 2019

@author: Mehdi zamani joharestani
"""

from math import sqrt
from scipy.interpolate import interpn
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import r2_score
from sklearn.model_selection import GridSearchCV
from sklearn.model_selection import train_test_split
import datetime
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pprint
import xgboost as xgb

# density_scatter plot function 
def density_scatter( x , y, ax = None, sort = True, bins = 20, figsize=(6,4), **kwargs )   :
    """
    Scatter plot colored by 2d histogram
    """
    x=x.reshape(len(x))
    y=y.reshape(len(y))
    if ax is None :
        fig , ax = plt.subplots(figsize=figsize)
    data , x_e, y_e = np.histogram2d( x, y, bins = bins)
    z = interpn( ( 0.5*(x_e[1:] + x_e[:-1]) , 0.5*(y_e[1:]+y_e[:-1]) ) , data , np.vstack([x,y]).T , method = "splinef2d", bounds_error = False )

    # Sort the points by density, so that the densest points are plotted last
    if sort :
        idx = z.argsort()
        x, y, z = x[idx], y[idx], z[idx]
    cmap = matplotlib.cm.get_cmap('jet')
    normalize = matplotlib.colors.Normalize(vmin=0, vmax=max(z))#min(z)
    ax.scatter( x, y, c=z, **kwargs ,cmap=cmap)
    cax, _ = matplotlib.colorbar.make_axes(ax)
    cbar = matplotlib.colorbar.ColorbarBase(cax, cmap=cmap, norm=normalize)
    z = np.polyfit(x, y, 1)
    p = np.poly1d(z)
    print(z)
    ax.plot(x,p(x),"r-")
    ax.plot(x,x,":",color='grey')
    return ax,fig,z
#======================================================================

# RMSE calculation function
def rmse(y_actual, y_predicted):
    return sqrt(mean_squared_error(y_actual, y_predicted))
#======================================================================

dfte= pd.DataFrame( columns=['features','r2','mae','rmse'])
#======================================================================
for file in ['totalnull.csv']: # name of databases 
    for step in ['NAOD','AOD03','AOD310','AOD10']:
        if step =='NAOD' :
            excludelist=[ 'AOD03', 'AOD10']
        elif step =='AOD03' :
            excludelist=[  'AOD10']
        elif step =='AOD10' :
            excludelist=[ 'AOD03']
        elif step =='AOD310' :
            excludelist=[]           
        hint=file[0:-4]+step
        df = pd.read_csv(file, na_values=['NA', ' '])
        df = df.drop(df.columns[df.columns.str.contains('Unnamed',case = False)],axis = 1) 
        lis=['Tehran_12','Fath square', 'Mahallati' ,'Tehran_15','Tehran_17' ]  # stations we want to exclude from analysis
        df=df[ ~df['Station'].isin(lis)]   
        df_=df[[
        #'date',
         'Dayofyear',
         'Weekday',
         'Season',
         'Latitude',
         'Longitude',
         'Altitude',
         'Station',
         'Organizati',
         'Distance',
         'T',
         'Tmax',
         'Tmin',
         'RH',
         'Rainfall',
         'Rainfall_Lag1',
         'Rainfall_Lag2',
         'visibility ',
         'wind_speed',
         'sustained_windsp',
         'Temperature',
         'Pressure',
         'Dewpoint',
         'Windspeed',
         'Winddir',
         'AOD03',
         'AOD10',
         'CO',
         'O3',
         'NO2',
         'SO2',
         'PM10',
         'PM2.5',
         'PM2.5_lag1',
         'PM2.5_lag2',
         'AQI'
        ]]
        
#======================================================================

# filtering the missing values based on features of experiment
        if step =='NAOD' :
            df_=df_[(df_[ 'PM2.5'] >= 0) &   (df_['PM2.5_lag1']>= 0) & (df_['PM2.5_lag2']>= 0)  ].reset_index(drop=True)
        elif step =='AOD03' :
            df_=df_[(df_[ 'PM2.5'] >= 0) &   (df_['PM2.5_lag1']>= 0) & (df_['PM2.5_lag2']>= 0) & (df_['AOD03']>= 0)  ].reset_index(drop=True)
        elif step =='AOD10' :
            df_=df_[(df_[ 'PM2.5'] >= 0) &   (df_['PM2.5_lag1']>= 0) & (df_['PM2.5_lag2']>= 0) & (df_['AOD10']>= 0)  ].reset_index(drop=True)
        elif step =='AOD310' :
            df_=df_[(df_[ 'PM2.5'] >= 0) &   (df_['PM2.5_lag1']>= 0) & (df_['PM2.5_lag2']>= 0) & (df_['AOD10']>= 0) & (df_['AOD03']>= 0)  ].reset_index(drop=True)      
        
# X=input features and Y=target selection    
        
        Ydataset=df_[[
         'PM2.5'
         ]]
        
        y= Ydataset.values
        
        
        Xdataset=df_[[
         'Dayofyear',
         'Weekday',
         'Season',
         'Latitude',
         'Longitude',
         'Altitude',
         'Organizati',
         'Distance',
         'T',
         'Tmax',
         'Tmin',
         'RH',
         'Rainfall',
         'Rainfall_Lag1',
         'Rainfall_Lag2',
         'visibility ',
         'wind_speed',
         'sustained_windsp',
         'Pressure',
         'Dewpoint',
         'AOD03',
         'AOD10',
         'PM2.5_lag1',
         'PM2.5_lag2',
        ]]

        x = Xdataset
        
#====================================================================== 


# train test date set prepration
        dataset_size=len(x)
        
        # train test data set preparation
        x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.30, random_state=3)
#======================================================================   

#Grid search regressor parameters configuration        
        start_time  = datetime.datetime.now()
        xgb_clf = xgb.XGBRegressor(objective = 'reg:linear') 
        
        # gride search parameters 
        parameters = {
                      'n_estimators': [200],     # any value in the form of  [ x,y,z,....]
                      'max_depth':[8],           # any value in the form of  [ x,y,z,....]
                      'gamma':[0.7],             # any value in the form of  [ x,y,z,....]
                      'min_child_weight':[8],    # any value in the form of  [ x,y,z,....]
                      }
        # grid search function 
        grid_search = GridSearchCV(estimator=xgb_clf, param_grid=parameters, cv=10, n_jobs=-1,verbose=42,scoring='r2')
#======================================================================         

# printing the results of grid search 
        
        print("parameters:")
        pprint.pprint(parameters)
        grid_search.fit(x_train, y_train)
        print("Best score: %0.3f" % grid_search.best_score_)
        print("Best parameters set:")
        best_parameters=grid_search.best_estimator_.get_params()
        for param_name in sorted(parameters.keys()):
            print("\t%s: %r" % (param_name, best_parameters[param_name]))
        end_time = datetime.datetime.now()
        print ('Select Done..., Time Cost: %d' % ((end_time - start_time).seconds))        
        print(grid_search.feature_importances_)
#======================================================================            
# evaluting the model using test dataset
        y_predxgb = grid_search.predict(x_test)
        mae=mean_absolute_error(y_test, y_predxgb)
        r2=r2_score( y_test, y_predxgb)
        rmse_value= rmse(y_test, y_predxgb)
        print(' R2  %.3f= '  % r2 )
        print(' MAE %.3f= '  % mae)
        print(' RMSE  %.3f= '  % rmse_value )
        print(r2,'  ',mae,' ', rmse_value)
        data=[hint, r2, mae, rmse_value] 
        dfte=dfte.append(pd.DataFrame(  [data]  , columns=['features','r2','mae','rmse']))
        text='R2= '+format(r2, '.2f') + ', MAE= '+format(mae, '.2f')+ ', RMSE= '+format(rmse_value, '.2f')
#====================================================================== 

# ploting density scatter graph
        ax,fig1,z=density_scatter( y_test.reshape(len(y_test)), y_predxgb, bins = [15,15] , figsize=(10,8))
        pol_text='y=%.2f x+%.2f, DB size=%0.1fK' % (z[0],z[1],(dataset_size/1000))
        ax.set_title('XGBoost Regressor Modeling' , fontsize=16,fontweight='bold')
        ax.set_xlabel("Observed PM2.5", fontsize=16)
        ax.set_ylabel("Predicted PM2.5", fontsize=16)
        ax.legend([pol_text,'y=x ',text], fontsize='x-large')
        ax.tick_params(axis='both', which='major', labelsize=12)
        ax.patch.set_facecolor('white')
        ax.set_ylim(0,250)
        ax.set_xlim(0,250)
#====================================================================== 
        
# ploting feature importance
        importances = grid_search.best_estimator_.feature_importances_
# Sort feature importances in descending order
        indices = np.argsort(importances)[::-1]
        featuersname= list(x.columns)
# Rearrange feature names so they match the sorted feature importances
        names = [featuersname[i] for i in indices]
# Barplot: Add bars      
        fig = plt.figure(figsize=(10,6))
        plt.style.use('bmh')
        ax = plt.subplot("111")
        ax.bar(range(x_train.shape[1]), importances[indices])
        # Add feature names as x-axis labels
        plt.xticks(range(x_train.shape[1]), names, rotation=90)
        # Create plot title
        ax.set_title('XGBoost regressor feature importance')
        plt.xlabel("Features")
        plt.ylabel("Score")
#======================================================================       
# save the loops results

dfte.to_csv('./xgb_result.csv')
        
